# Distribution Checker
# Common Test Module (Test_common.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


# Common interface class for all test objects.
	
package Test_common;
use strict;

use IO::Handle;

use Misc;
use Packages;
use Download;
use Report;
use Manifest;
use Subshell;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = (
	@Report::EXPORT,      # re-export Report::*
	@Packages::EXPORT,    # re-export Packages::*
	@Manifest::EXPORT,    # re-export Manifest::*
);

# Class constructor
sub new {
	my ($class, %init_data) = @_;
	
	my $self  = {};
	$self->{STATE} = 'Not started';

	for my $key ( keys %init_data ) {
		$self->{$key} = $init_data{$key};
	}

	bless $self, $class;
	return $self;
}

# Read methods for class fields
sub name ()     { return $_[0]->{NAME} };
sub temp_dir () { return $_[0]->{'temp_dir'} };
sub journal_file () { return $globals->{'test_result_dir'}."/".$_[0]->{JOURNAL_FILE} };

sub state {
	my ($self, $newval) = @_;
	
	if ( defined $newval ) {
		( $newval =~ /^(Not started|Preparing|Running|Making report|Finished|Crashed)$/ )
			or warning "Wrong STATE value: '$newval'";
		my $oldval = $self->{STATE};
		$self->{STATE} = $newval;
		if ( $newval eq 'Preparing' ) {
			$self->{PREPARE_TIME} = time();
		}
		elsif ( $newval eq 'Running' ) {
			$self->{START_TIME} = time() if !defined $self->{START_TIME};
			
			# ESTIM_START_TIME is used for estimating time in the progress() method.
			$self->{ESTIM_START_TIME} = $self->{START_TIME};
		}
		elsif ( $newval eq 'Making report' || $newval eq 'Finished' || $newval eq 'Crashed' ) {
			$self->{FINISH_TIME} = time() if !defined $self->{FINISH_TIME};
		}
		is_ok status_updated()
			or warning "Failed to update status", $Error::Last;
		return $oldval;
	} else {
		return $self->{STATE};
	}
}

#----------------------------------------------------------------------

# Initial checks.
sub check {
	my ($self) = @_;
	
	return 1; # OK
}

# Prepare the test.
sub prepare {
	my ($self) = @_;
	
	# Download and install required packages
	is_ok $self->prepare_test_packages()  # According to the Manifest
		or return error "Failed to download and install packages for '"
					.$self->name."' test.", $Error::Last;
	
	return 1; # OK
}

# Run the test
sub run {
	my ($self) = @_;
	
	return error ref($self)."->run() method should be overrided!";
}

sub after_run {
	my ($self) = @_;
	
	return;
}

sub make_report {
	my ($self) = @_;
	
	if ( !$self->{JOURNAL_FILE} ) {
		return error("No journal.");
	}
	if ( !$self->{JOURNAL_TYPE} ) {
		$self->report_warning(warning "Journal type wasn't specified. Trying to guess it.");
		$self->{JOURNAL_TYPE} = guess_journal_type($self->{JOURNAL_FILE});
	}
	
	my $file = $globals->{'test_result_dir'}."/".$self->{JOURNAL_FILE};
	
	if ( !-f $file ) {
		return error "No journal file: '$file'.";
	}
	
	my $jstruct = {
		FILE => $file,
		KIND => $self->{JOURNAL_TYPE},
		TESTSUITE => $self->name,
	};
	$jstruct->{TESTS_EXPECTED} = $self->{TESTS_EXPECTED} if defined $self->{TESTS_EXPECTED};
	
	# Print a link to the journal
	$self->report( '<p>Open the test '.Report::local_href( $file, 'journal' ).'.</p>' );
	
	is_ok Report::ParseJournal( $jstruct )
		or return error "Failed to parse journal: '$file'.", $Error::Last;
	
	# Print the errors table
	$self->report( Report::JournalReport($jstruct) );
	
	# Print the statistics
	$self->report( Report::print_stats($jstruct->{STATS}) );

	$self->{REPORT_STATS} = $jstruct->{STATS};
	$self->{VERDICT} = Report::verdict( $jstruct->{STATS} );
	
	return 1;
}

sub Cleanup {
	my ($self) = @_;
	
	return;
}

#----------------------------------------------------------------------

sub set_opt_defaults {
	my ($self) = @_;
	
	return if !defined $self->{OPTIONS};
	
	# Set all undefined options to default values
	foreach my $opt ( values %{$self->{OPTIONS}} ) {
		if ( !defined $opt->{VALUE} ) {
			if ( defined $opt->{DEFAULT} ) {
				$opt->{VALUE} = $opt->{DEFAULT};
			}
		}
	}
}

#----------------------------------------------------------------------

sub Go {
	my ($self) = @_;
	
	# Open the report file
	is_ok $self->open_report_file()
		or return error "Failed to open report file", $Error::Last;
	
	# Print title to the report
	$self->report( '<p><i>'.html($self->{DESCRIPTION}).'</i></p>' ) if $self->{DESCRIPTION};
	
	# Put a link to the current position in console log.
	if ( defined $globals->{LINE_N} ) {
		$self->report( '<p>' );
		my $line = $globals->{LINE_N}{0};
		$self->report( 'Open the '
					.Report::local_href( "log?line=${line}#line${line}", 'console log' ) ) if $line;
		if ( $line = $globals->{LINE_N}{1} ) {
			$self->report( ', ' );
			$self->report( Report::local_href( "verbose_log?line=${line}#line${line}", 'verbose log', 'target="verbose_log"' ) );
		}
		$self->report( '.</p>' );
	}
	
	# The important thing is that finish() MUST be called even if prepare() or run() fails.
	# So we wrap the major part into the run_part subroutine.
	
	local * run_part = sub {
		$self->{STARTED} = 1;
		$self->state('Preparing');
		
		# Check
		inform "Check ".$self->name." tests v. ".$self->{VERSION};
		
		my $res = eval{ $self->check() };
		if ( $@ ) { $res = error $@; }
		if ( is_err($res) || !$res ) {
			$res = error "The ".$self->name()." check failed.", $Error::Last;
			return $res unless $globals->{'ignore_check'};
			# else
			$self->report_error($res);
		}
		
		# Prepare
		
		inform "Prepare ".$self->name." tests v. ".$self->{VERSION};
		
		$res = eval{ $self->prepare() };
		if ( $@ ) { $res = error $@; }
		if ( is_err($res) || !$res ) {
			return error "Failed to prepare ".$self->name."."
					." The test will be skipped", $Error::Last;
		}
		
		# Non-changed options should be undefined in the prepare(),
		# because the undefined option value may be set there depending on global settings.
		# See Cmdchk_test.pm for example.
		
		# After the prepare() call we set all undefined options to its' default values.
		$self->set_opt_defaults();
		
		# Save the test configuration
		is_ok $self->save_info()
			or return error "Failed to save the test configuration ".$self->name."."
					." The test will be skipped", $Error::Last;
		
		# Run
		
		inform "Run ".$self->name." tests v. ".$self->{VERSION};
		
		$self->state('Running');
		
		$res = eval{ $self->run() }; # run
		if ( $@ ) { $res = error $@; }
		if ( is_err($res) || !$res ) {
			return error $self->name." test has crashed.", $Error::Last;
		}
	};
	
	my $run_res = run_part(); # ^^^

	$self->{FINISH_TIME} = time();
	
	# If failed
	if ( is_err($run_res) ) {
		$self->report_error(complain $Error::Last);
	}
	
	# Finish
	
	my $res = eval{ $self->finish() };
	if ( $@ ) { $res = error $@; }
	if ( is_err($res) ) { warning $res; }
	
	# Cleanup
	is_ok $self->Cleanup() and $self->{CLEANUP_OK} = 1;
	
	if ( $self->{RUN_PROBLEMS} && defined $self->{SUBSHELL} ) {
		# print last lines from subshell to the report
		$self->report_error( $self->{SUBSHELL}->get_last_lines() );
	}
	
	# Save the test configuration
	is_ok $self->save_info()
		or complain "Failed to save the test configuration", $Error::Last;
	
	if ( $self->{REPORT_FILEHANDLE} ) {
		# Close the report file
		close $self->{REPORT_FILEHANDLE};
		$self->{REPORT_FILEHANDLE} = undef;
	}
	
	$self->state('Making report');
	$self->Report(); # Uses .report.part2
	
	if ( is_err($run_res) ) {
		$self->state('Crashed');
		return error $self->name." test run failed"; # Failed
	} else {
		$self->state('Finished'); # Finished successfully
		return 1; # Ok
	}
}

#----------------------------------------------------------------------

# This procedure is called even if the test is failed to run.
sub finish {
	my ($self) = @_;
	
	inform "Finishing ".$self->name."...";
	
	# Check the subshell
	$self->Check_subshell_off();
	
	if ( !$self->{START_TIME} ) {
		$self->report_warning(warning "The test was skipped.");
	}
	
	if ( $self->{START_TIME} ) { 
	# The test at least has started
	
		# Processing the results
		if ( !is_ok $self->after_run() ) {
			$self->report_error(complain $Error::Last);
		} else {
			inform "...done";
		}
	}
	
	return 1; # OK
}
#======================================================================

## Auxiliary functions

# Create a temporary directory
sub create_temp_dir {
	my ($self) = @_;
	
	( $globals->{'temp_dir'} ) or return error "Global temp directory isn't defined";
	
	my $temp_dir = $globals->{'temp_dir'}."/".$self->name;
	
	is_ok create_empty_directory( $temp_dir ) or return $Error::Last;
	
	$self->{'temp_dir'} = $temp_dir;

	return 1; # OK
}
#----------------------------------------------------------------------

sub open_report_file {
	my ($self) = @_;
	
	return 1 if $self->{REPORT_FILEHANDLE};
	
	my $report_file = $self->{REPORT_FILE};
	# Open file for writing
	if ( !defined $report_file ) {
		$report_file = $self->{REPORT_FILE}
			= $globals->{'test_result_dir'}."/".$self->name.".report.part1";
	}
	
	my $report_file_handle;
	open $report_file_handle, ">".$report_file
		or return error "Failed to open '$report_file' for writing: $!.\nProbably you should run this as root.";
	
	$report_file_handle->autoflush(1);
	
	$self->{REPORT_FILEHANDLE} = $report_file_handle;
	
	return 1;
}

sub report {
	my ($self, $txt) = @_;
	
	return unless defined $txt;
	
	return if !$self->{REPORT_FILEHANDLE};
	
	print { $self->{REPORT_FILEHANDLE} } $txt;
}

sub report_warning {
	my ($self, $err) = @_;
	
	# Don't set RUN_PROBLEMS status
	
	if ( !is_err($err) ) {
		if ( !defined $err || $err eq "" ) {
			return;
		} 
		$err = error $err;
	}
	
	# Print some error details to the report
	local $Error::Debug = 0;
	$self->report( Report::ErrMessage($err, 'Warning: ') );
}

sub report_error {
	my ($self, $err) = @_;
	
	$self->{RUN_PROBLEMS} = 1;
	
	goto &report_warning;
}
#----------------------------------------------------------------------

# Link to the subshell log
sub report_sslog_link {
	my ($self) = @_;
	
	if ( !$self->{'sslogfile'} ) {
		return error "Subshell log file is undefined.";
	}
	if ( !-f $globals->{'test_result_dir'}."/".$self->{'sslogfile'} ) {
		return error "No subshell output log: '"
				.$globals->{'test_result_dir'}."/".$self->{'sslogfile'}."'.";
	}
	
	# Don't print the same link twice
	return 1 if $self->{SSLOG_LINKS}{$self->{'sslogfile'}};
	$self->{SSLOG_LINKS}{$self->{'sslogfile'}} = 1;
	
	$self->report( '<p>Open the '
				.Report::local_href( "results/".$self->{'sslogfile'}.'?line=0', 'subshell log' )
			.'.</p>' );
	
	return 1; # OK
}

#----------------------------------------------------------------------

sub Report {
	my ($self) = @_;
	
	inform "Processing ".$self->name." ...";
	
	if ( $self->{REPORT_FILEHANDLE} ) {
		close $self->{REPORT_FILEHANDLE};
		$self->{REPORT_FILEHANDLE} = undef;
	}
	# Open new report file: test.report.part2
	$self->{REPORT_FILE} = $globals->{'test_result_dir'}."/".$self->name.".report.part2";
	is_ok $self->open_report_file()
		or fail $Error::Last;
	
	my $res = eval{ $self->make_report() };
	if ( $@ ) { $res = error $@; }
	is_ok($res) or $self->report_error(complain $res);
	
	if ( $self->{RUN_PROBLEMS} || !$self->{VERDICT} ) {
		$self->{VERDICT} = "run_problems";
	}
	
	# Close the report file
	if ( $self->{REPORT_FILEHANDLE} ) {
		close $self->{REPORT_FILEHANDLE};
		$self->{REPORT_FILEHANDLE} = undef;
	}
	
	$self->save_info();
}

#----------------------------------------------------------------------

sub save_info {
	my ($self) = @_;
	
	my $filename = $globals->{'test_result_dir'}."/".$self->name.".info";
	
	my $res = "";
	
	local *save = sub {
			my ($name, $value) = @_;
			
			$value =~ s/\n/\x81/g;
			
			$res .= "$name: $value\n";
		};
	
	# Save all scalar fields
	foreach my $key ( sort keys %$self ) {
		my $value = $self->{$key};
		next if ref($value) ne "";
		next if !defined $value;
		
		save($key, $value);
	}
	
	# Save files
	if ( defined $self->{REQUIREMENTS} && @{$self->{REQUIREMENTS}} ) {
		if ( defined $self->{FILES} && %{$self->{FILES}} ) {
			my @files = ();
			for my $req ( @{$self->{REQUIREMENTS}} ) {
				my $file = $self->{FILES}{$req->{PACKAGE}} or next;
				push @files, $file->{LOCATION}."/".$file->{FILENAME};
			}
			save( 'FILES_', join(",", @files) );
		}
		if ( defined $self->{PACKAGES} && %{$self->{PACKAGES}} ) {
			my @packages = ();
			for my $req ( @{$self->{REQUIREMENTS}} ) {
				my $ver = ($self->{PACKAGES}{$req->{PACKAGE}} or "");
				push @packages, $req->{PACKAGE}."-".$ver;
			}
			save( 'PACKAGES_', join(",", @packages) );
		}
	}
	# Save options
	if ( $self->{OPTIONS} ) {
		foreach my $ID ( sort keys %{$self->{OPTIONS}} ) {
			my $opt = $self->{OPTIONS}{$ID};
			
			my $value = $opt->{VALUE};
			if ( !defined $value ) { $value = "(undef)"; }
			
			save( "OPT_".$ID, $value );
		}
	}
	# Save REPORT_STATS
	if ( $self->{REPORT_STATS} ) {
		foreach my $key ( sort keys %{$self->{REPORT_STATS}} ) {
			my $value = $self->{REPORT_STATS}{$key};
			( defined $value ) or $value = "";
			
			save( "REPORT_STAT_".$key, $value );
		}
	}
	
	is_ok write_string_as_file( $filename, $res ) or return $Error::Last;
	
	return 1;
}
#----------------------------------------------------------------------

# Download and install packages required by a test
sub prepare_test_packages {
	my ($self, $check_only) = @_;
	
	my $test_name = $self->{NAME};
	
	my $manifest = $globals->{'manifest'}
		or return error "Manifest wasn't read.";
	
	my $standard = $globals->{'standard'};
	
	my $architecture = $host_info->{'architecture'};
	my $package_manager = $host_info->{'package_manager'};

	my $test_ver_mnf = $self->{VERSION_MNF} or return error "Test version hasn't been selected.";
	my $status = $self->{STATUS} or return error "Test version status wasn't set.";

	my $requirements = $manifest->{COMPOSITION}{$standard}{$test_name}{$architecture}{$status}{$test_ver_mnf}
		or return error "Requirements aren't specified for"
					." '$standard' - '$test_name' '$test_ver_mnf', '$architecture', status '$status';";

	$self->{REQUIREMENTS} = $requirements;
	$self->{FILES} = {}; # $self->{FILES}{$requirements->{PACKAGE}} = $file;
	$self->{PACKAGES} = {}; # $self->{FILES}{$requirements->{PACKAGE}} = $version;

	foreach my $req ( @$requirements ) {
		my $package_name = $req->{PACKAGE};
		my $package_ver  = $req->{PACKAGE_VER}; # May be undef!
		my $location     = $req->{LOCATION}; # May be undef!

		my $req_tags = {};
		if ( defined $req->{TAG} ) {
			for my $tag ( split /;/, $req->{TAG} ) {
				$req_tags->{$tag} = 1;  # TODO: testme (AUTOUNPACK)
			}
		}

		if ( !$location ) {
			# Just require the package to be installed
			my $installed_ver = package_info( $package_name, $architecture );
			if ( !$installed_ver ) {
				return error "Package '$package_name'".($package_ver ? " v. '$package_ver'":"")." required by '$test_name' isn't installed.";
			}
			else {
				if ( !$package_ver ) { # No version requirement
					inform "Package '$package_name' is installed.";
				} else {
					if ( !version_filter($package_ver, $installed_ver) ) {
						return error "Test '$test_name' requires '$package_name' v. '$package_ver' while version '$installed_ver' is installed.";
					} else {
						if ( $installed_ver eq $package_ver ) { 
							inform "Package '$package_name' v. '$package_ver' is installed.";
						} else {
							inform "Package '$package_name' v. '$package_ver' ($installed_ver) is installed.";
						}
					}
				}
			}

			$self->{PACKAGES}{$package_name} = $installed_ver;
			next;
		}
		
		###
		my $found_files = Manifest::select_file($manifest, $architecture, $package_manager, $req);
		if ( is_err($found_files) || !$found_files || !@$found_files ) {
			return error "select_files: Can't find file '$package_name'"
				.($package_ver ? " v. '$package_ver'" : "")
				." ($package_manager), '$standard', arch: '$architecture',"
				." required by $test_name"
				.($location ? " in '$location'" : ""), $Error::Last;
		}
		if ( @$found_files > 1 ) {
			warning "Ambigous file choise ($package_name".($package_ver ? " v. '$package_ver'" : "")."): ".join ", ", map {$_->{SRC}} @$found_files;
		}
		
		my $file = $found_files->[0]; # Take any one
		$self->{FILES}{$package_name} = $file;

		###

		my $filename = $file->{FILENAME};
		my $dest_dir = cache_dest( $file->{SRC} );
		my $file_in_cache = $dest_dir."/".$filename;

		# Look at the file on server
		my $file_was_rewrited = 0;
		unless ( $file->{OPTIONS}{KEEP} ) {
			if ( $Download::enabled ) {
				next if $check_only;
				my $link;
				my $sp = {}; # Information about the file on server
				my $mirror_i = 0;
				MIRROR_LOOP: {
					# Try available mirrors
					$link = file_link($globals->{'manifest'}, $file, $globals->{'dl_method'}, $mirror_i++);
					if ( is_err($link) ) {
						last MIRROR_LOOP if $Error::Last->{-text} eq 'No more mirrors';
						return error "Can't construct an URL for file '".$file->{SRC}."/".$file->{FILENAME}."'"
							."\nThere may be errors in the Manifest.", $Error::Last;
					}

					# Spider query
					$sp = spider($link);
					last MIRROR_LOOP if ( $sp->{SIZE} ); # ok

					# else
					if ( $sp->{NOTFOUND} ) {
						inform "Can't download '$link':\n The file was removed or the server is unavailable.";
					} else {
						inform "Can't access '$link'.";
					}
					redo MIRROR_LOOP;
				}
				if ( !$sp->{SIZE} ) {
					# If haven't found a good mirror then take the first one.
					$link = file_link($globals->{'manifest'}, $file, $globals->{'dl_method'});
				}

				# Check if we needed to download the file.
				my $need_download = eval {
					if ( !-f $file_in_cache ) { # The file hasn't been downloaded
						return 1;
					}
					if ( $file->{OPTIONS}{RELOAD} ) {
						debug_inform "It was chosen to re-download file '$filename'.";
						return 1;
					}
					# Check the file
					my ($size, $mtime) = file_info($file_in_cache);
					if ( defined $sp->{SIZE} && $sp->{SIZE} != $size ) {
						debug_inform "File size differs for '$filename': $sp->{SIZE} != $size.";
						return 1;
					}
					if ( defined $sp->{TIMESTAMP} && $sp->{TIMESTAMP} > $mtime ) { # The file is newer
						debug_inform "File on server is newer: '$filename': $sp->{TIMESTAMP} > $mtime.";
						return 1;
					}
					return 0;
				};

				if ( $need_download ) {
					if ( $sp->{TIMESTAMP} && $file->{TIMESTAMP} ) {
						# Check the timestamp
						if ( $sp->{TIMESTAMP} < $file->{TIMESTAMP} ) {
							# TODO: fail if certification?
							warning "The file at the server is older than specified in the Manifest. ($sp->{TIMESTAMP} < $file->{TIMESTAMP}). This should not happen.";
						}
					}
					# Fix the $sp if needed
					if ( !defined $sp->{TIMESTAMP} ) {
						$sp->{TIMESTAMP} = $file->{TIMESTAMP}; # Trust the manifest, though it is not a reliable source.
					}

					# Download the file
					is_ok download_file( $link, $file_in_cache, RESUME=>1, SPIDER=>$sp, OVERWRITE=>1 )
						or return error "Failed to download file '$filename'", $Error::Last;

					$file_was_rewrited = 1;
				}
				else {
					inform "File '$filename' is already downloaded.";
				}
			}
			
			# Check the downloaded file
			if ( ! -f $file_in_cache ) {
				if ( !$Download::enabled ) {
					return error "Downloading files from the Internet was not enabled.";
				} else {
					return error "There is no file '$filename' required by '$test_name' (probably failed to download).";
				}
			}
			# Check the file
			{
				my ($size, $mtime) = file_info($file_in_cache);
				
				# Check the timestamp
				if ( $file->{TIMESTAMP} && $mtime < $file->{TIMESTAMP} ) {
					# TODO: fail if certification?
					warning "File '$filename' is older than specified in the Manifest. ($mtime < $file->{TIMESTAMP}). This should not happen.";
				}
				if ( $file->{SIZE} && $size != $file->{SIZE} ) {
					debug_inform "Size of file '$filename' differs from one in the Manifest. ($size != $file->{SIZE}). This may happen with snapshot files.";
				}
			}
		} else {
			debug_inform("Don't check file '$filename' due to user's choice.");
		}

		next if $check_only;

		if ( $file->{OPTIONS}{RELOAD} ) {
			# It was chosen to re-download this file.
			# Reset this flag to avoid downloading this file again for another test.
			$file->{OPTIONS}{RELOAD} = 0;
		}

		if ( $file->{KIND} eq 'rpm' || $file->{KIND} eq 'deb' ) {
			# Install the package

			my $installed_ver;
			my $need_install = 1;
			my @deinstall = (); # packages to be deinstalled before installing the required package.

			# Check if the package is already installed
			$installed_ver = package_info( $package_name, $architecture );
			
			if ( defined $installed_ver ) { # installed
				if ( !$package_ver || version_filter($package_ver, $installed_ver) ) {
					# Correct version is already installed
					$need_install = 0;
					# but:
					if ( $globals->{'force_reinstall'} ) {
						$need_install = 1;
						inform "Forced reinstallation...";
					}
					elsif ( $file->{OPTIONS}{REINSTALL} ) {
						$need_install = 1;
						inform "Forced reinstallation for '$package_name' '$installed_ver'.";
					}
					elsif ( $file_was_rewrited ) {
						$need_install = 1;
						inform "A newer version of the file '$filename' was downloaded, force its installation.";
					}
				} else { # Another version of the package is installed.
					$need_install = 1;
				}
			} else { # The package hasn't been installed
				$need_install = 1;
			}
			
			if ( $need_install ) {
				if ( defined $installed_ver ) { # already installed
					# Uninstall it.
					push @deinstall, [$package_name, $installed_ver];
				}
			}
			else {
				if ( defined $installed_ver ) { # already installed (should be TRUE)
					inform "Package '$package_name' '$installed_ver' is already installed.";
				}
				next;
			}

			# First, deinstall conflicting packages
			if ( defined $manifest->{CONFLICTS} ) {
				if ( defined $manifest->{CONFLICTS}{$package_name} ) {
					foreach my $package ( keys %{$manifest->{CONFLICTS}{$package_name}} ) {
						my $ver = package_info( $package, $architecture );
						if ( defined $ver ) { # a conflicting package is installed
							push @deinstall, [$package, $ver];
						}
					}
				}
			}
			for my $elem ( @deinstall ) {
				my ($package, $ver) = @$elem;

				# TODO: fix: the package architecture should be taken into consideration!
				if ( glob_hash($manifest->{PACKAGES}{$package}, 'KEEP', $ver) ) {
					# It was chosen by user to keep this package installed
					return error "Package '$package' '$ver': cannot deinstall due to user's choise.";
				}
				
				remove_package( $package ); # Ignore the result.
			}
			
			# Install the package
			$last_cmd_output = undef;
			is_ok install_package( $file_in_cache, $globals->{'temp_dir'} )
				or return error "Failed to install '$package_name' package"
								." required by $test_name test.", $Error::Last;
			my $installer_output = $last_cmd_output;
			
			# Check the result
			$installed_ver = package_info( $package_name, $architecture );
			if ( !defined $installed_ver # not installed
				|| ($package_ver && !version_filter($package_ver, $installed_ver)) # wrong version
			) {
				my $installer_err = undef;
				$installer_err = error "Installer's output:\n".$installer_output if defined $installer_output;
				return error "Failed to install '$package_name'".($package_ver ? " v. '$package_ver'" : "")." package required by $test_name test.", $installer_err;
			}
			
			inform "Package '$package_name' was installed successfully";
			next;
		}
		elsif ( $file->{KIND} eq 'file' ) {
			if ( $req_tags->{AUTOUNPACK} ) {
				my $test_unpack_dir = $VAR_OPT_DIR.'/data/unpack/'.$test_name;
				# Remove directory, then create it clean and unpack the file.
				if ( -d $test_unpack_dir ) {
					cmd('rm -rf '.shq($test_unpack_dir));
				}
				cmd('mkdir -p '.shq($test_unpack_dir));
				( -d $test_unpack_dir ) or return error "Failed to create directory '$test_unpack_dir'";
				
				is_ok unpack_file($file_in_cache, $test_unpack_dir)
					or return $Error::Last;
			}
			next;
		}
	}

	return 1;
}

# Creates and prepares a subshell
sub Spawn_subshell {
	my ($self, $user, $sslog) = @_;
	
	$sslog = $self->name.".sslog" if !defined $sslog;
	$self->{'sslogfile'} = $sslog;
	my $sslog_file = $globals->{'test_result_dir'}."/".$sslog;
	
	my $subshell = Subshell::New();
	is_ok $subshell->LogFile( $sslog_file )
		or return error "Failed to open subshell log file", $Error::Last;
	
	is_ok $subshell->Spawn($user) # TODO: compare all tests with $user=undef vs. $user='root'
		or return error "Failed to spawn a subshell", $Error::Last;
	
	$self->{SUBSHELL} = $subshell;
	
	# Put a link to the subshell log into the report.
	$self->report_sslog_link();
	
	$subshell->Settle() or warning "Too much noise on a just started subshell";
	is_ok $subshell->Send("export LC_ALL=C; export LC_MESSAGES=C; export LC_COLLATE=C; export LC_CTYPE=C; export LC_MONETARY=C; export LC_NUMERIC=C; export LC_TIME=C;\n") # To avoid messages in local languages
		or return error "Failed to send command to the started subshell", $Error::Last;

	return $subshell;
}

sub Check_subshell_off {
	my ($self) = @_;
	
	# Check the subshell is closed
	if ( $self->{SUBSHELL} && !$self->{SUBSHELL}{TERMINATED} ) {
		my $pid_info = ( $self->{SUBSHELL}{PTY_PID} ? "PTYSHELL: ".$self->{SUBSHELL}{PTY_PID} : "")
				.( $self->{SUBSHELL}{SUB_PID} ? ", SUBSHELL: ".$self->{SUBSHELL}{SUB_PID} : "");
				
		inform "Terminating ".$self->name." (".$pid_info.")";
		if ( !is_ok $self->{SUBSHELL}->ExpectLogout(0) ) {
			$self->report_error( complain "Attention! Failed to shutdown a subshell."
					."\nPlease, check the process list and kill the following processes: ".$pid_info,
				$Error::Last );
		} else {
			$self->report_error(complain "The test was terminated.");
		}
		$self->state('Crashed');
		
		# Save the test configuration
		is_ok $self->save_info()
			or complain "Failed to save the test configuration", $Error::Last;
		
		return 1;
	}
}
#----------------------------------------------------------------------

# Copy file. 
sub copy_file {
	my ($self, $file, $copy_to) = @_;
	
	if ( !-e $file ) {
		return error "File doesn't exist: '$file'.";
	}
	
	cmd("cp -rf ".shq($file)." ".shq($copy_to)) == 0
		or return error "Failed to copy: '$file' to '$copy_to'.";
	
	return 1;
}


# Allows "globbed" source.
sub copy_file_globbed {
	my ($self, $glob, $copy_to) = @_;
	
	my @files = glob( $glob ) # Get files
		or return error "Can't find file: '$glob'";
	
	my $last_file = pop @files
		or return error "Can't find any files: '$glob'";

	if ( @files ) { # Unexpected files
		warning "These files are ignored: @files";
	}
	
	is_ok $self->copy_file ($last_file, $copy_to)
		or return $Error::Last;
	
	return 1;
}

#----------------------------------------------------------------------

# Copies one of required files (by name) to specified directory.
sub copy_files_from_reqs_to {
	my ($self, $dest_dir) = @_;

	( defined $self->{FILES} && %{$self->{FILES}} )
		or return error "There are no files selected for test ".$self->name;
	
	foreach my $file ( values %{$self->{FILES}} ) {
		next if $file->{KIND} ne 'file';
		my $file_in_the_cache = cache_dest($file->{SRC})."/".$file->{FILENAME};
		# Check the file
		( -f $file_in_the_cache )
			or return error "File '".$file->{FILENAME}."' hasn't been downloaded"
					." ('$file_in_the_cache' expected).";
		
		# Copy the file from cache to the directory specifed
		cmd( "cp -f ".shq($file_in_the_cache)." ".shq($dest_dir) );
		
		# Check the result of the unpacking
		( -f "$dest_dir/".$file->{FILENAME} )
			or return error "Failed to copy '$file_in_the_cache' to '$dest_dir'";
	}
	
	return 1;
}
#-----------------------------------------------------------------------

sub option {
	my ($self, $optname, $equals) = @_;
	
	return undef if !defined $self->{OPTIONS};
	return undef if !defined $self->{OPTIONS}{$optname};
	
	my $value = $self->{OPTIONS}{$optname}{VALUE};
	if ( !defined $value ) {
		$value = $self->{OPTIONS}{$optname}{DEFAULT};
	}
	return undef if !defined $value;
	
	if ( defined $equals ) {
		return ($value eq $equals);
	}
	return $value;
}
#-----------------------------------------------------------------------

sub progress {
	my ($self, $ratio) = @_;
	
	( defined $ratio ) or return error "Undefined ratio";
	
	my $time = time();
	my $elapsed = $time - $self->{ESTIM_START_TIME};
	
	if ( $ratio > 1 ) { $ratio = 1; }
	my $prepare_time = $self->{ESTIM_START_TIME} - $self->{PREPARE_TIME};
	$self->{ESTIM_TIME} = $prepare_time + int($elapsed/$ratio) if $ratio > 0;
	if ( !defined $self->{PREPARE_PERCENT} ) {
		$self->{PREPARE_PERCENT} = $prepare_time / $self->{ESTIM_TIME} * 100;
	}
	$self->{PERCENT} = int( $ratio*100*(1-$self->{PREPARE_PERCENT}/100) + $self->{PREPARE_PERCENT});
	if ( DEBUG ) { myprint "progress: ".$self->{PERCENT}."\n", 3; }
	
	$self->{LAST_PROGRESS_UPDATE} = $time;
	
	status_updated();
}

sub progress_profiled {
	my ($self, $checkpoints, $testcase) = @_;
	
	if ( defined $testcase ) {
		# Look for the $testcase in the $checkpoints table
		my $found = undef;
		for( my $i = 0; $i < @$checkpoints; $i++ ) {
			my $pattern = $checkpoints->[$i][1];
			
			if ( $testcase eq $pattern ) {
				$found = $i;
				last;
			}
		}
		if ( $found ) {
			my $ratio = $checkpoints->[$found][0] / 100;
			if ( DEBUG ) { myprint "checkpoint: ".int($ratio*100)."\n", 3; }
			my $next;
			if ( defined $checkpoints->[$found+1] ) {
				$self->{PROGRESS_NEXT} = $checkpoints->[$found+1][0] / 100;
			} else {
				$self->{PROGRESS_NEXT} = 1;
			}
			
			$self->progress($ratio);
			return;
		}
	}
	
	# else: average the progress
	
	if ( defined $self->{ESTIM_TIME} && $self->{ESTIM_TIME} > 1 ) {
		my $time = time();
		my $elapsed = $time - $self->{ESTIM_START_TIME};
		my $ratio = $elapsed / $self->{ESTIM_TIME};
		if ( $ratio >= 1 ) { $ratio = 0.99; }
		return if ( int($ratio*100) <= $self->{PERCENT} );
		return if ( defined $self->{PROGRESS_NEXT} && $ratio > $self->{PROGRESS_NEXT} );
		
		$self->progress($ratio);
	}
}

#-----------------------------------------------------------------------
#-----------------------------------------------------------------------
1; # return value
